@@ -41,7 +41,6 @@ import os
 import fnmatch
 import glob
 import shutil
-import codecs
 import hashlib
 import errno
 import tempfile
@@ -78,24 +77,15 @@ string_types = str,
 text_type = str
 getcwdu = os.getcwd
 
-def surrogate_escape(error):
-    """
-    Simulate the Python 3 ``surrogateescape`` handler, but for Python 2 only.
-    """
-    chars = error.object[error.start:error.end]
-    assert len(chars) == 1
-    val = ord(chars)
-    val += 0xdc00
-    return __builtin__.unichr(val), error.end
 
 if PY2:
     import __builtin__
     string_types = __builtin__.basestring,
     text_type = __builtin__.unicode
     getcwdu = os.getcwdu
-    codecs.register_error('surrogateescape', surrogate_escape)
     map = itertools.imap
 
+
 @contextlib.contextmanager
 def io_error_compat():
     try:
@@ -109,6 +99,7 @@ def io_error_compat():
 
 ##############################################################################
 
+
 __all__ = ['Path', 'CaseInsensitivePattern']
 
 
@@ -133,7 +124,7 @@ class TreeWalkWarning(Warning):
 
 # from jaraco.functools
 def compose(*funcs):
-    compose_two = lambda f1, f2: lambda *args, **kwargs: f1(f2(*args, **kwargs))
+    compose_two = lambda f1, f2: lambda *args, **kwargs: f1(f2(*args, **kwargs))  # noqa
     return functools.reduce(compose_two, funcs)
 
 
@@ -325,23 +316,45 @@ class Path(text_type):
         return self.expandvars().expanduser().normpath()
 
     @property
-    def namebase(self):
+    def stem(self):
         """ The same as :meth:`name`, but with one file extension stripped off.
 
-        For example,
-        ``Path('/home/guido/python.tar.gz').name == 'python.tar.gz'``,
-        but
-        ``Path('/home/guido/python.tar.gz').namebase == 'python.tar'``.
+        >>> Path('/home/guido/python.tar.gz').stem
+        'python.tar'
         """
         base, ext = self.module.splitext(self.name)
         return base
 
+    @property
+    def namebase(self):
+        warnings.warn("Use .stem instead of .namebase", DeprecationWarning)
+        return self.stem
+
     @property
     def ext(self):
         """ The file extension, for example ``'.py'``. """
         f, ext = self.module.splitext(self)
         return ext
 
+    def with_suffix(self, suffix):
+        """ Return a new path with the file suffix changed (or added, if none)
+
+        >>> Path('/home/guido/python.tar.gz').with_suffix(".foo")
+        Path('/home/guido/python.tar.foo')
+
+        >>> Path('python').with_suffix('.zip')
+        Path('python.zip')
+
+        >>> Path('filename.ext').with_suffix('zip')
+        Traceback (most recent call last):
+        ...
+        ValueError: Invalid suffix 'zip'
+        """
+        if not suffix.startswith('.'):
+            raise ValueError("Invalid suffix {suffix!r}".format(**locals()))
+
+        return self.stripext() + suffix
+
     @property
     def drive(self):
         """ The drive specifier, for example ``'C:'``.
@@ -432,8 +445,9 @@ class Path(text_type):
     @multimethod
     def joinpath(cls, first, *others):
         """
-        Join first to zero or more :class:`Path` components, adding a separator
-        character (:samp:`{first}.module.sep`) if needed.  Returns a new instance of
+        Join first to zero or more :class:`Path` components,
+        adding a separator character (:samp:`{first}.module.sep`)
+        if needed.  Returns a new instance of
         :samp:`{first}._next_class`.
 
         .. seealso:: :func:`os.path.join`
@@ -678,7 +692,7 @@ class Path(text_type):
             try:
                 isfile = child.isfile()
                 isdir = not isfile and child.isdir()
-            except:
+            except Exception:
                 if errors == 'ignore':
                     continue
                 elif errors == 'warn':
@@ -705,8 +719,8 @@ class Path(text_type):
             attribute, it is applied to the name and path prior to comparison.
 
         `normcase` - (optional) A function used to normalize the pattern and
-            filename before matching. Defaults to :meth:`self.module`, which defaults
-            to :meth:`os.path.normcase`.
+            filename before matching. Defaults to :meth:`self.module`, which
+            defaults to :meth:`os.path.normcase`.
 
         .. seealso:: :func:`fnmatch.fnmatch`
         """
@@ -926,14 +940,15 @@ class Path(text_type):
             to read the file later.
         """
         with self.open('ab' if append else 'wb') as f:
-            for l in lines:
-                isUnicode = isinstance(l, text_type)
+            for line in lines:
+                isUnicode = isinstance(line, text_type)
                 if linesep is not None:
                     pattern = U_NL_END if isUnicode else NL_END
-                    l = pattern.sub('', l) + linesep
+                    line = pattern.sub('', line) + linesep
                 if isUnicode:
-                    l = l.encode(encoding or sys.getdefaultencoding(), errors)
-                f.write(l)
+                    line = line.encode(
+                        encoding or sys.getdefaultencoding(), errors)
+                f.write(line)
 
     def read_md5(self):
         """ Calculate the md5 hash for this file.
@@ -947,8 +962,8 @@ class Path(text_type):
     def _hash(self, hash_name):
         """ Returns a hash object for the file at the current path.
 
-            `hash_name` should be a hash algo name (such as ``'md5'`` or ``'sha1'``)
-            that's available in the :mod:`hashlib` module.
+        `hash_name` should be a hash algo name (such as ``'md5'``
+        or ``'sha1'``) that's available in the :mod:`hashlib` module.
         """
         m = hashlib.new(hash_name)
         for chunk in self.chunks(8192, mode="rb"):
@@ -1171,7 +1186,8 @@ class Path(text_type):
                 gid = grp.getgrnam(gid).gr_gid
             os.chown(self, uid, gid)
         else:
-            raise NotImplementedError("Ownership not available on this platform.")
+            msg = "Ownership not available on this platform."
+            raise NotImplementedError(msg)
         return self
 
     def rename(self, new):
@@ -1231,7 +1247,8 @@ class Path(text_type):
             self.rmdir()
         except OSError:
             _, e, _ = sys.exc_info()
-            if e.errno != errno.ENOTEMPTY and e.errno != errno.EEXIST:
+            bypass_codes = errno.ENOTEMPTY, errno.EEXIST, errno.ENOENT
+            if e.errno not in bypass_codes:
                 raise
         return self
 
@@ -1301,11 +1318,16 @@ class Path(text_type):
             return self._next_class(newpath)
 
     if hasattr(os, 'symlink'):
-        def symlink(self, newlink):
+        def symlink(self, newlink=None):
             """ Create a symbolic link at `newlink`, pointing here.
 
+            If newlink is not supplied, the symbolic link will assume
+            the name self.basename(), creating the link in the cwd.
+
             .. seealso:: :func:`os.symlink`
             """
+            if newlink is None:
+                newlink = self.basename()
             os.symlink(self, newlink)
             return self._next_class(newlink)
 
@@ -1409,19 +1431,23 @@ class Path(text_type):
     # in-place re-writing, courtesy of Martijn Pieters
     # http://www.zopatista.com/python/2013/11/26/inplace-file-rewriting/
     @contextlib.contextmanager
-    def in_place(self, mode='r', buffering=-1, encoding=None, errors=None,
-            newline=None, backup_extension=None):
+    def in_place(
+            self, mode='r', buffering=-1, encoding=None, errors=None,
+            newline=None, backup_extension=None,
+    ):
         """
-        A context in which a file may be re-written in-place with new content.
+        A context in which a file may be re-written in-place with
+        new content.
 
-        Yields a tuple of :samp:`({readable}, {writable})` file objects, where `writable`
-        replaces `readable`.
+        Yields a tuple of :samp:`({readable}, {writable})` file
+        objects, where `writable` replaces `readable`.
 
         If an exception occurs, the old file is restored, removing the
         written data.
 
-        Mode *must not* use ``'w'``, ``'a'``, or ``'+'``; only read-only-modes are
-        allowed. A :exc:`ValueError` is raised on invalid modes.
+        Mode *must not* use ``'w'``, ``'a'``, or ``'+'``; only
+        read-only-modes are allowed. A :exc:`ValueError` is raised
+        on invalid modes.
 
         For example, to add line numbers to a file::
 
@@ -1447,22 +1473,28 @@ class Path(text_type):
         except os.error:
             pass
         os.rename(self, backup_fn)
-        readable = io.open(backup_fn, mode, buffering=buffering,
-            encoding=encoding, errors=errors, newline=newline)
+        readable = io.open(
+            backup_fn, mode, buffering=buffering,
+            encoding=encoding, errors=errors, newline=newline,
+        )
         try:
             perm = os.fstat(readable.fileno()).st_mode
         except OSError:
-            writable = open(self, 'w' + mode.replace('r', ''),
+            writable = open(
+                self, 'w' + mode.replace('r', ''),
                 buffering=buffering, encoding=encoding, errors=errors,
-                newline=newline)
+                newline=newline,
+            )
         else:
             os_mode = os.O_CREAT | os.O_WRONLY | os.O_TRUNC
             if hasattr(os, 'O_BINARY'):
                 os_mode |= os.O_BINARY
             fd = os.open(self, os_mode, perm)
-            writable = io.open(fd, "w" + mode.replace('r', ''),
+            writable = io.open(
+                fd, "w" + mode.replace('r', ''),
                 buffering=buffering, encoding=encoding, errors=errors,
-                newline=newline)
+                newline=newline,
+            )
             try:
                 if hasattr(os, 'chmod'):
                     os.chmod(self, perm)
@@ -1585,8 +1617,9 @@ class Multi:
 
 class tempdir(Path):
     """
-    A temporary directory via :func:`tempfile.mkdtemp`, and constructed with the
-    same parameters that you can use as a context manager.
+    A temporary directory via :func:`tempfile.mkdtemp`, and
+    constructed with the same parameters that you can use
+    as a context manager.
 
     Example:
 
@@ -1625,7 +1658,8 @@ def _multi_permission_mask(mode):
     >>> _multi_permission_mask('a=r,u+w')(0) == 0o644
     True
     """
-    compose = lambda f, g: lambda *args, **kwargs: g(f(*args, **kwargs))
+    def compose(f, g):
+        return lambda *args, **kwargs: g(f(*args, **kwargs))
     return functools.reduce(compose, map(_permission_mask, mode.split(',')))
 
 
@@ -1707,3 +1741,189 @@ class CaseInsensitivePattern(text_type):
     @property
     def normcase(self):
         return __import__('ntpath').normcase
+
+
+class FastPath(Path):
+    """
+    Performance optimized version of Path for use
+    on embedded platforms and other systems with limited
+    CPU. See #115 and #116 for background.
+    """
+
+    def listdir(self, pattern=None):
+        children = os.listdir(self)
+        if pattern is None:
+            return [self / child for child in children]
+
+        pattern, normcase = self.__prepare(pattern)
+        return [
+            self / child
+            for child in children
+            if self._next_class(child).__fnmatch(pattern, normcase)
+        ]
+
+    def walk(self, pattern=None, errors='strict'):
+        class Handlers:
+            def strict(msg):
+                raise
+
+            def warn(msg):
+                warnings.warn(msg, TreeWalkWarning)
+
+            def ignore(msg):
+                pass
+
+        if not callable(errors) and errors not in vars(Handlers):
+            raise ValueError("invalid errors parameter")
+        errors = vars(Handlers).get(errors, errors)
+
+        if pattern:
+            pattern, normcase = self.__prepare(pattern)
+        else:
+            normcase = None
+
+        return self.__walk(pattern, normcase, errors)
+
+    def __walk(self, pattern, normcase, errors):
+        """ Prepared version of walk """
+        try:
+            childList = self.listdir()
+        except Exception:
+            exc = sys.exc_info()[1]
+            tmpl = "Unable to list directory '%(self)s': %(exc)s"
+            msg = tmpl % locals()
+            errors(msg)
+            return
+
+        for child in childList:
+            if pattern is None or child.__fnmatch(pattern, normcase):
+                yield child
+            try:
+                isdir = child.isdir()
+            except Exception:
+                exc = sys.exc_info()[1]
+                tmpl = "Unable to access '%(child)s': %(exc)s"
+                msg = tmpl % locals()
+                errors(msg)
+                isdir = False
+
+            if isdir:
+                for item in child.__walk(pattern, normcase, errors):
+                    yield item
+
+    def walkdirs(self, pattern=None, errors='strict'):
+        if errors not in ('strict', 'warn', 'ignore'):
+            raise ValueError("invalid errors parameter")
+
+        if pattern:
+            pattern, normcase = self.__prepare(pattern)
+        else:
+            normcase = None
+
+        return self.__walkdirs(pattern, normcase, errors)
+
+    def __walkdirs(self, pattern, normcase, errors):
+        """ Prepared version of walkdirs """
+        try:
+            dirs = self.dirs()
+        except Exception:
+            if errors == 'ignore':
+                return
+            elif errors == 'warn':
+                warnings.warn(
+                    "Unable to list directory '%s': %s"
+                    % (self, sys.exc_info()[1]),
+                    TreeWalkWarning)
+                return
+            else:
+                raise
+
+        for child in dirs:
+            if pattern is None or child.__fnmatch(pattern, normcase):
+                yield child
+            for subsubdir in child.__walkdirs(pattern, normcase, errors):
+                yield subsubdir
+
+    def walkfiles(self, pattern=None, errors='strict'):
+        if errors not in ('strict', 'warn', 'ignore'):
+            raise ValueError("invalid errors parameter")
+
+        if pattern:
+            pattern, normcase = self.__prepare(pattern)
+        else:
+            normcase = None
+
+        return self.__walkfiles(pattern, normcase, errors)
+
+    def __walkfiles(self, pattern, normcase, errors):
+        """ Prepared version of walkfiles """
+        try:
+            childList = self.listdir()
+        except Exception:
+            if errors == 'ignore':
+                return
+            elif errors == 'warn':
+                warnings.warn(
+                    "Unable to list directory '%s': %s"
+                    % (self, sys.exc_info()[1]),
+                    TreeWalkWarning)
+                return
+            else:
+                raise
+
+        for child in childList:
+            try:
+                isfile = child.isfile()
+                isdir = not isfile and child.isdir()
+            except Exception:
+                if errors == 'ignore':
+                    continue
+                elif errors == 'warn':
+                    warnings.warn(
+                        "Unable to access '%s': %s"
+                        % (self, sys.exc_info()[1]),
+                        TreeWalkWarning)
+                    continue
+                else:
+                    raise
+
+            if isfile:
+                if pattern is None or child.__fnmatch(pattern, normcase):
+                    yield child
+            elif isdir:
+                for f in child.__walkfiles(pattern, normcase, errors):
+                    yield f
+
+    def __fnmatch(self, pattern, normcase):
+        """ Return ``True`` if `self.name` matches the given `pattern`,
+        prepared version.
+        `pattern` - A filename pattern with wildcards,
+            for example ``'*.py'``. The pattern is expected to be normcase'd
+            already.
+        `normcase` - A function used to normalize the pattern and
+            filename before matching.
+        .. seealso:: :func:`Path.fnmatch`
+        """
+        return fnmatch.fnmatchcase(normcase(self.name), pattern)
+
+    def __prepare(self, pattern, normcase=None):
+        """ Prepares a fmatch_pattern for use with ``FastPath.__fnmatch`.
+        `pattern` - A filename pattern with wildcards,
+            for example ``'*.py'``. If the pattern contains a `normcase`
+            attribute, it is applied to the name and path prior to comparison.
+        `normcase` - (optional) A function used to normalize the pattern and
+            filename before matching. Defaults to :meth:`self.module`,
+            which defaults to :meth:`os.path.normcase`.
+        .. seealso:: :func:`FastPath.__fnmatch`
+        """
+        if not normcase:
+            normcase = getattr(pattern, 'normcase', self.module.normcase)
+        pattern = normcase(pattern)
+        return pattern, normcase
+
+    def fnmatch(self, pattern, normcase=None):
+        if not pattern:
+            raise ValueError("No pattern provided")
+
+        pattern, normcase = self.__prepare(pattern, normcase)
+        return self.__fnmatch(pattern, normcase)
